#!/usr/bin/env python3
"""
This example illustrates how to reconstruct a multivariate density
by decomposing it into a product of the copula and the margnals 
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 7 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND, makeOrderedFromFlat, \
    CopulaHisto, fillCopulaHisto
from npstat_density_estimation import *
import matplotlib.pyplot as plt
from npstat_plotting import *

# Some parameters for the script
sampleSize = 1000    # Number of points to generate (sample size).
nDiscreteCopula = 50 # Number of discretization cells in each dimension
                     # to use for copula creation.
nDiscreteMarg = 512  # Number of discretization cells for the marginals.
symbetaPower = -1    # Power parameter of the symmetric beta kernel
                     # used to reconstruct the marginals. Specify -1
                     # to use a Gaussian kernel instead.
filterDegree = 2     # Polynomial degree used to generate high-order kernels
                     # for estimating the marginals.

# We will use a 2-d Gassian mixture to generate the random sample.
# The code below constructs such a mixture. The weights do not have
# to add up to 1, they will be renormalized internally.
mixture = ns.DistributionMixND(2)
center0 = (1.0, 0.5)
covarianceMatrix0 = ((2.0, 0.3),
                     (0.3, 0.5))
weight0 = 7
mixture.add(ns.GaussND(center0, covarianceMatrix0), weight0)
center1 = (-1.0, -0.5)
covarianceMatrix1 = ((2.0, -2.0),
                    (-2.0,  3.0))
weight1 = 3
mixture.add(ns.GaussND(center1, covarianceMatrix1), weight1)

# Generate a random sample according to this density.
# Note that the "generate" function of multivariate densities
# generates a flat list of doubles, not a multivariate sequence.
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram for visualizing the sample
xaxis = ns.HistoAxis(25, -5.0, 5.0, "X")
yaxis = ns.HistoAxis(25, -5.0, 5.0, "Y")
h0 = HistoND("int", xaxis, yaxis, "Sample Histogram", "Counts")
h0.fill(sample, 1)

# Histogram X and Y separately (the marginals)
hx = HistoND("int", xaxis, "X Marginal", "Counts")
sampleX = ns.makeFlatFromFlat(sample, mixture.dim(), (0,))
hx.fill(sampleX, 1)

hy = HistoND("int", yaxis, "Y Marginal", "Counts")
sampleY = ns.makeFlatFromFlat(sample, mixture.dim(), (1,))
hy.fill(sampleY, 1)

# Create the copula histogram. There is a helper function "CopulaHisto"
# for making these.
h1 = CopulaHisto(((nDiscreteCopula, "X"), (nDiscreteCopula, "Y")),
                 "Copula Histogram", "Counts")

# Convert flat sample into a sample of "Ordered Points". These objects
# can remember their place in an ordered sequence according to a number
# of different ordering criteria. Fill the copula histogram (a special
# method "fillCopulaHisto" is needed for this).
ordered = makeOrderedFromFlat(mixture.dim(), sample)
fillCopulaHisto(ordered, h1)

# Create the object which will perform cross-validation calculations
cvCalc = BandwidthCVPseudoLogliND('double', 0.5)

# Create the object which will smooth the copula. In this example,
# we will smooth it with tensor product of Bernstein polynomials.
# Bernstein polynomials have a special "partition of unity" property
# which helps in building copula smoothers. The polynomial degree
# will be the same in each direction. This is the property of the
# output of the "bernsteinCopulaSmootherDegrees" function, but other
# meaningful strategies can be designed as well. Another good method
# for smoothing copulas is LOrPE (class ns.LOrPECopulaSmoother).
smoother = ns.BernsteinCopulaSmoother(
    [nDiscreteCopula]*mixture.dim(), 1.0e-8, 1000, cvCalc, False,
    bernsteinCopulaSmootherDegrees(mixture.dim(), 2, nDiscreteCopula))

# Smooth the copula. The first argument of the "smooth" function
# is a unique dataset id. In this example it is ignored.
bestEffectiveBandwidth = smoother.smooth(0, ordered)

# Convert the effective bandwidth returned into the chosen
# degree of Bernstein polynomials. Note that this only works if
# "bernsteinCopulaSmootherDegrees" function was used to specify
# the degrees.
bestDeg = bernsteinCopulaSmootherBwToDeg(bestEffectiveBandwidth)
print("Bernstein polynomial degree chosen by cross-validation is", bestDeg)

# Get the smoothed copula (produced during the last "smooth" call)
smoothedCopula = smoother.getHisto()

# Procedure for reconstructing the marginals. In this example,
# we will use FFT-based KDE and RLCV for choosing the bandwidth.
# Note that, if you plan to reconstruct many distributions, you
# don't want to recreate the smoother and repopulate it with
# kernels every time, as it is done below.
def estimateMarginalDensity(label, h, filterDegree, symbetaPower):
    smoother = FFTKDE1D(h, False)
    bw0 = histoPluginBandwidth1D(h, h.nFillsTotal(),
                                 filterDegree, symbetaPower)
    for bw in ns.EquidistantInLogSpace(bw0/2.0, bw0*2.0, 50):
        smoother.addFilter((filterDegree, bw, symbetaPower))
    rlcvCalc = BinnedRLCVCalc1D(0.5)
    bestRLCV, paramsRLCV, yRLCV = smoother.bestFilterByCV(h, rlcvCalc)
    distro = ns.BinnedDensity1D(h.axis(0).min(), h.axis(0).length(), yRLCV, 1)
    rlcvFilterDegree, rlcvBandwidth, rlcvSymbetaPower = paramsRLCV
    print("{} marginal plug-in bandwidth is {:.4f}, "\
          "RLCV bandwidth is {:.4f}".format(label, bw0, rlcvBandwidth))
    return h.axis(0).binCenters(), yRLCV, distro

# Estimate the density of the marginals
mx = HistoND("int", xaxis.rebin(nDiscreteMarg), "Discretized X", "Counts")
mx.fill(sampleX, 1)
binCentersX, rlcvScanX, xmarg = estimateMarginalDensity(
    "X", mx, filterDegree, symbetaPower)

my = HistoND("int", yaxis.rebin(nDiscreteMarg), "Discretized Y", "Counts")
my.fill(sampleY, 1)
binCentersY, rlcvScanY, ymarg = estimateMarginalDensity(
    "Y", my, filterDegree, symbetaPower)

# Combine the marginals and the copula into an overall multivariate
# distribution
copulaDistro = ns.BinnedDensityND(smoothedCopula.lowerAxisLimits(),
                                  smoothedCopula.axisLengths(),
                                  smoothedCopula.binContents(), 1)
multivariateDistro = ns.CompositeDistributionND(copulaDistro, xmarg, ymarg)

# Scan the true density for subsequent display
box = h0.boundingBox()
scanShape = (nDiscreteMarg, nDiscreteMarg)
trueDensity = ns.scanDensityND(mixture, box, scanShape)

# Scan the estimated distribution for subsequent display
estimatedDensity = ns.scanDensityND(multivariateDistro, box, scanShape)

# Calculate the integrated squared error (ISE)
cellArea = box.volume()/nDiscreteMarg/nDiscreteMarg
ise = ns.sumOfSquaredDifferences(trueDensity, estimatedDensity)*cellArea
print("ISE is {:.5f}".format(ise))

# Plot various things
fig, axs = plt.subplots(4, 2,figsize=(6, 8))
colormeshNumpyArray(axs[0, 0], trueDensity, box,
                    flipy=True, title="True Density")
colormeshHisto2D(axs[0, 1], h0)
colormeshHisto2D(axs[1, 0], h1)
colormeshHisto2D(axs[1, 1], smoothedCopula, title="Smoothed Copula")
colormeshNumpyArray(axs[2, 0], estimatedDensity, box,
                    flipy=True, title="Estimated Density")
colormeshNumpyArray(axs[2, 1], estimatedDensity - trueDensity, box,
                    flipy=True, title="Estimated - True")
plotHistoOutline1D(axs[3, 0], hx)
axs[3, 0].plot(binCentersX, rlcvScanX*hx.integral(), 'r')
plotHistoOutline1D(axs[3, 1], hy)
axs[3, 1].plot(binCentersY, rlcvScanY*hy.integral(), 'r')

# Display the window
fig.suptitle("Density Estimation by Copula Decomposition")
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
