#!/usr/bin/env python3
"""
This example illustrates how to perform orthogonal series density
estimation (OSDE) in 1-d with discretized data
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 20 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import DiscreteOSDE1D, \
                                      BinnedLSCVCalc1D, BinnedRLCVCalc1D
import matplotlib.pyplot as plt
from npstat_plotting import plotHistoOutline1D

# Some parameters for the script
sampleSize = 500     # Number of points to generate (sample size)
xmin = -3.5          # Left edge of the sample histogram
xmax = 3.5           # Right edge of the sample histogram
numBins = 70         # Number of bins for the sample histogram
nDiscrete = 700      # Number of cells for sample discretization
allowNegative = True # Specify whether the density estimate is allowed
                     # to be negative somewhere

# We will use a 1-d Gassian mixture to generate the random sample.
# See the "brute_force_kde_1d.py" script for more comments on this.
component1 = ns.GaussianMixtureEntry(3.0, -1.0, 0.5)
component2 = ns.GaussianMixtureEntry(7.0,  1.0, 0.5)
mixture = ns.GaussianMixture1D(0.0, 1.0, (component1, component2))

# Generate a random sample according to this density
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram for visualizing the sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Histogram for sample discretization
h2 = HistoND("int", ns.HistoAxis(nDiscrete, xmin, xmax, "X"),
             "OSDE Discretization Histogram", "Counts")
h2.fill(sample, 1)

# Object which will perform OSDE
smoother = DiscreteOSDE1D(h2, allowNegative)

# Initial density estimate. Maximum polynomial degree is estimated
# by the Hart scheme.
degHart = smoother.optimalDegreeHart(h2)
yHart = smoother.estimateDataDensity(degHart)
print("Hart scheme degree is", degHart)

# Add a number of filters with different degrees for cross-validation
# calculations. Note that these degrees do not have to be integer.
for degree in ns.EquidistantInLinearSpace(2.0, smoother.maxDegree(), 100):
    smoother.addFilter(degree)

# Run the least squares cross-validation (LSCV)
lscvCalc = BinnedLSCVCalc1D()
bestLSCV, lscvFilterDegree, yLSCV = smoother.bestFilterByCV(h2, lscvCalc)
print("Best LSCV degree is", lscvFilterDegree)

# Run the regularized likelihood cross-validation (RLCV)
regularisationParameter = 0.5
rlcvCalc = BinnedRLCVCalc1D(regularisationParameter)
bestRLCV, rlcvFilterDegree, yRLCV = smoother.bestFilterByCV(h2, rlcvCalc)
print("Best RLCV degree is", rlcvFilterDegree)

# We are done with density estimation at this point. The rest of the script
# is just plotting results. To begin with, create a Matplotlib plot window.
fig = plt.figure()
ax = fig.add_subplot(111)

# Plot the sample histogram
plotHistoOutline1D(ax, h1, ylabel='Density estimate',
                   title='Legendre Basis OSDE')

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
verticalScale = h1.binVolume()*sampleSize
xdens, ydens = ns.scanDensity1D(mixture, xmin, xmax, nDiscrete+1)
ax.plot(xdens, ydens*verticalScale, 'g', label='True density')

# Overlay the OSDE results with various degrees
xcoords = h2.axis(0).binCenters()
colors     = ('c',           'r',     'b'   )
labels     = ('Hart scheme', 'RLCV',  'LSCV')
results    = (yHart,         yRLCV,   yLSCV )
for color, l, y in zip(colors, labels, results):
    ax.plot(xcoords, y*verticalScale, color, label=l)

# Add plot legend
plt.legend(loc=2)

# Display the window
plt.show()
