#!/usr/bin/env python3
"""
This example illustrates how to use a multivariate LOrPE filter
constructed as a tensor product of 1-d filters
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 30 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import SequentialLOrPE, \
    normalBandwidthFactor, scanMultivariateDensityAsWeight, \
    BinnedRLCVCalcND, BinnedLSCVCalcND
import matplotlib.pyplot as plt
from npstat_plotting import *

# Some parameters for the script
sampleSize = 1000    # Number of points to generate (sample size)
nDiscrete = 140      # Number of discretization cells to use (here, in each
                     # dimension). Should be a power of 2 for best FFT
                     # performance.
symbetaPower = -1    # Power parameter of the symmetric beta kernels.
                     # Specify -1 to use a Gaussian kernel instead.
filterDegree = 0     # Polynomial degree used to generate high-order kernels.
                     # In general, even numbers make the most sense here.
                     # Kernel order will be filterDegree + 2.
allowNegative = True # Specify whether the density estimate is allowed
                     # to be negative somewhere (this is possible in case
                     # filterDegree > 0)

# We will use a 2-d Gassian mixture to generate the random sample.
# The code below constructs such a mixture. The weights do not have
# to add up to 1, they will be renormalized internally.
mixture = ns.DistributionMixND(2)
center0 = (1.0, 0.5)
covarianceMatrix0 = ((2.0, 0.3),
                     (0.3, 0.5))
weight0 = 7
mixture.add(ns.GaussND(center0, covarianceMatrix0), weight0)
center1 = (-1.0, -0.5)
covarianceMatrix1 = ((2.0, -2.0),
                    (-2.0,  3.0))
weight1 = 3
mixture.add(ns.GaussND(center1, covarianceMatrix1), weight1)

# Generate a random sample according to this density.
# Note that the "generate" function of multivariate densities
# generates a flat list of doubles, not a multivariate sequence.
# The normal "fill" method of the HistoND class will work
# in such cases as well, consuming several coordinate values
# at a time. The number of elelements in the list should be
# divisible by the histogram dimensionality.
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram for visualizing the sample
xaxis = ns.HistoAxis(25, -5.0, 5.0, "X")
yaxis = ns.HistoAxis(25, -5.0, 5.0, "Y")
h1 = HistoND("int", xaxis, yaxis, "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Histogram for sample discretization. Note that the axis range is
# increased, to catch all possible tails. In general, such an increase
# may or may not make sense, depending on situation.
h2 = HistoND("int", ns.HistoAxis(nDiscrete, -7.0, 7.0, "X"),
                    ns.HistoAxis(nDiscrete, -7.0, 7.0, "Y"),
                    "KDE Discretization Histogram", "Counts")
h2.fill(sample, 1)

# Object which will perform the KDE
smoother = SequentialLOrPE(h2, allowNegative)

# Make an initial guess for the filter bandwidth. This guess will use
# the sample covariance matrix estimated from the histogram. Ideally,
# here we need some robust covariance matrix estimate instead -- minimum
# volume ellipsoid (MVE), minimum covariance determinant (MCD), etc.
pluginFactor = normalBandwidthFactor(h2.dim(), filterDegree, sampleSize)
pluginCovariance = pluginFactor*ns.histoCovariance(h2)
bwx0, bwy0 = np.sqrt(pluginCovariance.mainDiagonal())

# Build the density estimate using the plug-in kernel
pluginSpecs = ((filterDegree, bwx0, symbetaPower),
               (filterDegree, bwy0, symbetaPower))
pluginEstimate = smoother.estimateDensity(h2, pluginSpecs)

# Add a number of kernels with different bandwidth values
# for cross-validation calculations. In general, filterDegree and
# symbetaPower parameters can be simultaneously optimized as well.
for bwFactor in ns.EquidistantInLogSpace(0.4, 2.0, 20):
    filterSpecs = ((filterDegree, bwFactor*bwx0, symbetaPower),
                   (filterDegree, bwFactor*bwy0, symbetaPower))
    smoother.addFilter(filterSpecs)

# Run the least squares cross-validation (LSCV)
lscvCalc = BinnedLSCVCalcND()
bestLSCV, paramsLSCV, scanLSCV = smoother.bestFilterByCV(h2, lscvCalc)
bwx, bwy = [item[1] for item in paramsLSCV]
print("Best LSCV bandwidth factor is {:.4f}".format(bwx/bwx0))

# Run the regularized likelihood cross-validation (RLCV).
# For the description of this method, see section 4.2 in
# https://doi.org/10.1080/10485252.2017.1371715.
regularisationParameter = 0.5
rlcvCalc = BinnedRLCVCalcND(regularisationParameter)
bestRLCV, paramsRLCV, scanRLCV = smoother.bestFilterByCV(h2, rlcvCalc)
bwx, bwy = [item[1] for item in paramsRLCV]
print("Best RLCV bandwidth factor is {:.4f}".format(bwx/bwx0))

# Calculate the integrated squared error (ISE) for various estimates
box = h2.boundingBox()
trueDensity = ns.scanDensityND(mixture, box, h2.shape())
for e, tag in zip((pluginEstimate, scanLSCV, scanRLCV),
                  ("Plug-in",      "LSCV",    "RLCV" )):
    ise = ns.sumOfSquaredDifferences(trueDensity, e)*h2.binVolume()
    print(tag, "estimate ISE is {:.5f}".format(ise))

# Plot various things
fig, axs = plt.subplots(2, 3)
colormeshNumpyArray(axs[0, 0], trueDensity, box,
                    flipy=True, title="True Density")
colormeshHisto2D(axs[0, 1], h1)
colormeshNumpyArray(axs[0, 2], pluginEstimate, box,
                    flipy=True, title="TP LOrPE Plug-in")
colormeshNumpyArray(axs[1, 0], scanLSCV, box,
                    flipy=True, title="TP LOrPE LSCV")
colormeshNumpyArray(axs[1, 1], scanRLCV, box,
                    flipy=True, title="TP LOrPE RLCV")
colormeshNumpyArray(axs[1, 2], scanRLCV - trueDensity, box,
                    flipy=True, title="RLCV - True")

# Set x and y ranges of the plots to that of the sample histogram
for ax in [axs[i, j] for i in range(2) for j in range(3)]:
    ax.set_xlim(h1.axis(0).min(), h1.axis(0).max())
    ax.set_ylim(h1.axis(1).min(), h1.axis(1).max())

# Display the window
fig.suptitle("Density Estimation by Tensor Product LOrPE")
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
