#!/usr/bin/env python3
"""
This example illustrates how to perform variable-bandwidth univariate
kernel density estimation (KDE) with NPStat
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 17 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
import matplotlib.pyplot as plt

# Some parameters for the script
sampleSize = 500    # Number of points to generate (sample size)
xmin = -3.5         # Left edge of the sample histogram
xmax = 3.5          # Right edge of the sample histogram
numBins = 70        # Number of bins for the sample histogram
nDiscrete = 1000    # Number of cells for sample discretization
symbetaPower = -1   # Power parameter of the symmetric beta kernel.
                    # Specify -1 to use a Gaussian kernel instead.
bwFactor = 0.5      # Pilot estimate bandwidth will be set to the product
                    # of the plug-in bandwidth and this factor. The choice
                    # of the factor is empirical here.

# We will use a 1-d Gassian mixture to generate the random sample.
# See the "brute_force_kde_1d.py" script for more comments on this.
component1 = ns.GaussianMixtureEntry(3.0, -1.0, 0.5)
component2 = ns.GaussianMixtureEntry(7.0,  1.0, 0.5)
mixture = ns.GaussianMixture1D(0.0, 1.0, (component1, component2))

# Generate a random sample according to this density
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram for displaying the sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Sample discretization for KDE. For smoothing purposes, the bin width
# should be substantially smaller than any physical scale of interest.
h2 = HistoND("int", ns.HistoAxis(nDiscrete, 2*xmin, 2*xmax, "X"),
             "KDE Discretization Histogram", "Counts")
h2.fill(sample, 1)

# Perform variable-bandwidth KDE. "simpleVariableBandwidthSmooth1D"
# is the simplest driver function to use. It is appropriate when the
# histogrammed sample consists of unweighted points. Function
# "weightedVariableBandwidthSmooth1D" can be used if the points are
# weighted (it requires the user to specify the effective sample size).
# Function "variableBandwidthSmooth1D" is the most flexible one but it
# requires the user to come up with his/her own pilot bandwidth.
# Note that Python signatures of these functions are generated by the
# "npstat/wrap/variableBandwidthSmooth1DWrap.hh" header rather than
# directly taken from "npstat/stat/variableBandwidthSmooth1D.hh".
ykde, bw = ns.simpleVariableBandwidthSmooth1D(h2, symbetaPower, bwFactor)
print("Pilot bandwidth is {:.4f}".format(bw))

# Create Matplotlib plot window
fig = plt.figure()
ax = fig.add_subplot(111)

# Plot the sample histogram
xh, yh = ns.histoOutline1D(h1)
ax.plot(xh, yh, 'k')

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
verticalScale = h1.binVolume()*sampleSize
xdens, ydens = ns.scanDensity1D(mixture, xmin, xmax, nDiscrete+1)
ax.plot(xdens, ydens*verticalScale, 'g', label='True density')

# Overlay the KDE result. Scale it to the area of the histogram.
ax.plot(h2.axis(0).binCenters(), ykde*verticalScale, 'r', label='KDE')

# Set the x range of the plot to that of the sample histogram
ax.set_xlim(h1.axis(0).min(), h1.axis(0).max())

# Add labels and comments
plt.xlabel('X')
plt.ylabel('Density')
plt.title('Variable-Bandwidth KDE')
plt.legend(loc=2)

# Display the window
plt.show()
