#ifndef NPSTAT_GAUSS1DQUADRATURE_HH_
#define NPSTAT_GAUSS1DQUADRATURE_HH_

/*!
// \file Gauss1DQuadrature.hh
//
// \brief Adaptation of Gauss-Hermite quadratures that works with Gauss1D
//        densities as weights
//
// Author: I. Volobouev
//
// November 2020
*/

#include "npstat/nm/GaussHermiteQuadrature.hh"

namespace npstat {
    class Gauss1DQuadrature
    {
    public:
        enum {interval_quadrature = 0};

        Gauss1DQuadrature(unsigned npoints, long double mu, long double sigma);

        inline unsigned npoints() const {return ghq_.npoints();}
        inline long double mean() const {return mean_;}
        inline long double sigma() const {return sigma_;}

        void getAllAbscissae(long double* abscissae, unsigned len) const;
        void getAllWeights(long double* weights, unsigned len) const;

        /** Perform the quadrature */
        template <typename FcnResult, typename FcnArg>
        inline long double integrate(const Functor1<FcnResult,FcnArg>& fcn) const
            {return ghq_.integrateProb(mean_, sigma_, fcn);}

        /** Check if the rule with the given number of points is supported */
        inline static bool isAllowed(const unsigned npoints)
            {return GaussHermiteQuadrature::isAllowed(npoints);}
        
        /** The complete set of allowed rules, in the increasing order */
        inline static std::vector<unsigned> allowedNPonts()
            {return GaussHermiteQuadrature::allowedNPonts();}

        /**
        // Minimum number of points, among the supported rules, which
        // integrates a polynomial with the given degree exactly (with
        // the appropriate weight). Returns 0 if the degree is out of range.
        */
        inline static unsigned minimalExactRule(const unsigned polyDegree)
            {return GaussHermiteQuadrature::minimalExactRule(polyDegree);}

    private:
        GaussHermiteQuadrature ghq_;
        long double mean_;
        long double sigma_;
    };
}

#endif // NPSTAT_GAUSS1DQUADRATURE_HH_
