#include <cmath>
#include <cfloat>

#include "geners/binaryIO.hh"

#include "npstat/stat/Distribution1DReader.hh"

namespace {
    class MomDensityFunctor1D : public npstat::Functor1<long double, long double>
    {
    public:
        inline MomDensityFunctor1D(const npstat::AbsDistribution1D& fcn,
                                   const long double center,
                                   const unsigned order)
            : fcn_(fcn), center_(center), order_(order) {}

        inline ~MomDensityFunctor1D() {}

        inline long double operator()(const long double& a) const
        {
            const long double delta = a - center_;
            long double mfactor;
            switch (order_)
            {
            case 0U:
                mfactor = 1.0L;
                break;
            case 1U:
                mfactor = delta;
                break;
            case 2U:
                mfactor = delta*delta;
                break;
            default:
                mfactor = powl(delta, order_);
                break;
            }
            return mfactor*fcn_.density(a);
        }

    private:
        MomDensityFunctor1D();

        const npstat::AbsDistribution1D& fcn_;
        const long double center_;
        const unsigned order_;
    };    
}

namespace npstat {
    bool AbsScalableDistribution1D::write(std::ostream& os) const
    {
        gs::write_pod(os, location_);
        gs::write_pod(os, scale_);
        return !os.fail();
    }

    bool AbsScalableDistribution1D::read(std::istream& is,
                                         double* location, double* scale)
    {
        gs::read_pod(is, location);
        gs::read_pod(is, scale);
        return !is.fail();
    }

    AbsDistribution1D* AbsDistribution1D::read(const gs::ClassId& id,
                                               std::istream& in)
    {
        return StaticDistribution1DReader::instance().read(id, in);
    }

    long double AbsDistribution1D::empiricalMoment(
        const long double center, const unsigned order,
        const unsigned nIntegrationPoints,
        const bool useFejerQuadrature) const
    {
        if (!nIntegrationPoints) throw std::invalid_argument(
            "In npstat::AbsDistribution1D::empiricalMoment: "
            "number of integration points must be positive");

        MomDensityFunctor1D fcn(*this, center, order);
        const double xmin = this->quantile(0.0);
        const double xmax = this->quantile(1.0);

        if (useFejerQuadrature)
        {
            FejerQuadrature quad(nIntegrationPoints);
            return quad.integrate(fcn, xmin, xmax);
        }
        else
        {
            if (!GaussLegendreQuadrature::isAllowed(nIntegrationPoints))
                throw std::invalid_argument(
                    "In npstat::AbsDistribution1D::empiricalMoment: "
                    "unsupported number of integration points");
            GaussLegendreQuadrature glq(nIntegrationPoints);
            return glq.integrate(fcn, xmin, xmax);
        }
    }
}
