#include <stdexcept>

#include "npstat/nm/FejerQuadrature.hh"
#include "npstat/stat/DensityOrthoPoly1D.hh"

namespace npstat {
    DensityOrthoPoly1D::DensityOrthoPoly1D(const AbsDistribution1D& distro,
                                           const unsigned maxDegree,
                                           const unsigned nIntegPoints,
                                           const OrthoPolyMethod m)
        : distro_(distro.clone()), wsum_(0.0L), maxdeg_(maxDegree)
    {
        typedef std::pair<double,double> MeasurePt;

        if (!nIntegPoints) throw std::invalid_argument(
            "In npstat::DensityOrthoPoly1D constructor: the number of "
            "integration points must be positive");
        const FejerQuadrature quad(nIntegPoints);
        const DensityFunctor1D weight(distro);
        const std::vector<MeasurePt>& measure = quad.weightedIntegrationPoints(
            weight, distro_->quantile(0.0), distro_->quantile(1.0));
        assert(measure.size() == nIntegPoints);
        for (unsigned i=0; i<nIntegPoints; ++i)
            wsum_ += measure[i].second;
        if (wsum_ <= 0.0L) throw std::invalid_argument(
            "In npstat::DensityOrthoPoly1D constructor: total weight "
            "is not positive");
        Private::calcRecurrenceCoeffs(measure, maxdeg_, wsum_, m, &rCoeffs_);
    }

    std::pair<long double,long double>
    DensityOrthoPoly1D::recurrenceCoeffs(const unsigned deg) const
    {
        const Private::Recurrence& r = rCoeffs_.at(deg);
        return std::pair<long double,long double>(r.alpha, r.sqrbeta);
    }

    DensityOrthoPoly1D::DensityOrthoPoly1D(const DensityOrthoPoly1D& r)
        : distro_(r.distro_->clone()),
          rCoeffs_(r.rCoeffs_),
          wsum_(r.wsum_),
          maxdeg_(r.maxdeg_)
    {
    }

    DensityOrthoPoly1D& DensityOrthoPoly1D::operator=(const DensityOrthoPoly1D& r)
    {
        if (&r != this)
        {
            distro_ = CPP11_auto_ptr<AbsDistribution1D>(r.distro_->clone());
            rCoeffs_ = r.rCoeffs_;
            wsum_ = r.wsum_;
            maxdeg_ = r.maxdeg_;
        }
        return *this;
    }
}
