#ifndef NPSTAT_DISTRIBUTIONMIXND_HH_
#define NPSTAT_DISTRIBUTIONMIXND_HH_

/*!
// \file DistributionMixND.hh
//
// \brief A mixture of multivariate statistical distributions
//
// Author: I. Volobouev
//
// September 2019
*/

#include "npstat/stat/AbsDistributionND.hh"

namespace npstat {
    class AbsRandomGenerator;

    class DistributionMixND : public AbsDistributionND
    {
    public:
        DistributionMixND(unsigned dim);

        DistributionMixND(const DistributionMixND&);
        DistributionMixND& operator=(const DistributionMixND&);

        /** "Virtual copy constructor" */
        inline virtual DistributionMixND* clone() const
            {return new DistributionMixND(*this);}

        virtual ~DistributionMixND();

        /** 
        // Add a component to the mixture. Weight must be positive.
        // All weights will be normalized internally so that their
        // sum is 1.
        */
        DistributionMixND& add(const AbsDistributionND& distro, double weight);

        /**
        // Set all weights. Number of weights provided should be equal
        // to the number of components. Weights will be normalized internally.
        */
        void setWeights(const double* weights, unsigned nWeights);

        /** Number of mixture components */
        inline unsigned nComponents() const {return entries_.size();}

        /** Get the mixture component with the given number */
        inline const AbsDistributionND& getComponent(const unsigned n) const
            {return *entries_.at(n);}

        /** Get the component weight */
        double getWeight(unsigned n) const;

        /** Probability density */
        virtual double density(const double* x, unsigned dim) const;

        virtual void unitMap(const double* rnd,
                             unsigned bufLen, double* x) const;

        inline virtual bool mappedByQuantiles() const {return false;}

        virtual unsigned random(AbsRandomGenerator& g,
                                double* x, unsigned lenX) const;
        //@{
        /** Prototype needed for I/O */
        inline virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream&) const;
        //@}

        static inline const char* classname() {return "npstat::DistributionMixND";}
        static inline unsigned version() {return 1;}
        static DistributionMixND* read(const gs::ClassId& id, std::istream&);

    protected:
        virtual bool isEqual(const AbsDistributionND&) const;

    private:
        void normalize();

        std::vector<const AbsDistributionND*> entries_;
        std::vector<double> weights_;
        std::vector<double> weightCdf_;
        long double wsum_;
        bool isNormalized_;
    };
}

#endif // NPSTAT_DISTRIBUTIONMIXND_HH_
