#include <algorithm>
#include <utility>
#include <stdexcept>
#include <cfloat>

#include "npstat/nm/LinearMapper1d.hh"

#include "npstat/stat/OSDE1D.hh"

namespace {
    void validate_sample_size(const std::string& where,
                              const double expectedSampleSize)
    {
        if (expectedSampleSize <= 0.0)
            throw std::invalid_argument(
                "In npstat::OSDE1D::" + where +
                ": expected sample size must be positive");
    }

    void require_identical_support(const std::string& where,
                                   const npstat::OSDE1D& osde,
                                   const npstat::AbsDistribution1D& distro)
    {
        if (osde.xmin() != distro.quantile(0.0) ||
            osde.xmax() != distro.quantile(1.0))
            throw std::invalid_argument(
                "In npstat::OSDE1D::" + where +
                ": incompatible density support");
    }

    class OSDE1DHlp1 : public npstat::Functor1<double, double>
    {
    public:
        inline OSDE1DHlp1(const npstat::OSDE1D& p, const unsigned d1)
            : poly_(p), deg1_(d1) {}

        inline double operator()(const double& x) const
            {return poly_.poly(deg1_, x)*poly_.weight(x);}

    private:
        const npstat::OSDE1D& poly_;
        unsigned deg1_;
    };

    class OSDE1DHlp2 : public npstat::Functor1<double, double>
    {
    public:
        inline OSDE1DHlp2(const npstat::OSDE1D& p,
                          const unsigned d1, const unsigned d2,
                          const double c1, const double c2)
            : poly_(p), deg1_(d1), deg2_(d2), c1_(c1), c2_(c2) {}

        inline double operator()(const double& x) const
        {
            const double w = poly_.weight(x);
            const std::pair<double,double>& p = poly_.twopoly(deg1_, deg2_, x);
            return (w*p.first - c1_)*(w*p.second - c2_);
        }

    private:
        const npstat::OSDE1D& poly_;
        unsigned deg1_, deg2_;
        double c1_, c2_;
    };
}

namespace npstat {
    OSDE1D::OSDE1D(const AbsClassicalOrthoPoly1D& poly1d,
                   const double xlo, const double xhi)
        : poly_(poly1d.clone()), xmin_(xlo), xmax_(xhi)
    {
        if (xmin_ == xmax_) throw std::invalid_argument(
            "In npstat::OSDE1D constructor: "
            "density estimation interval has zero length");
        if (xmin_ > xmax_)
            std::swap(xmin_, xmax_);
        const double polyWidth = poly_->xmax() - poly_->xmin();
        scale_ = (xmax_ - xmin_)/polyWidth;
        shift_ = (poly_->xmax()*xmin_ - poly_->xmin()*xmax_)/polyWidth;
    }

    OSDE1D::OSDE1D(const OSDE1D& r)
        : poly_(r.poly_->clone()),
          xmin_(r.xmin_), xmax_(r.xmax_),
          shift_(r.shift_), scale_(r.scale_)
    {
    }

    OSDE1D& OSDE1D::operator=(const OSDE1D& r)
    {
        if (&r != this)
        {
            delete poly_; poly_ = 0;
            poly_ = r.poly_->clone();
            xmin_ = r.xmin_;
            xmax_ = r.xmax_;
            shift_ = r.shift_;
            scale_ = r.scale_;
        }
        return *this;
    }

    OSDE1D::~OSDE1D()
    {
        delete poly_;
    }

    std::pair<double,double> OSDE1D::twopoly(
        const unsigned deg1, const unsigned deg2, const double x) const
    {
        const std::pair<long double,long double>& p =
            poly_->twopoly(deg1, deg2, (x - shift_)/scale_);
        return std::pair<double,double>(p.first, p.second);
    }

    double OSDE1D::series(const double *coeffs, const unsigned maxdeg,
                          const double x) const
    {
        return poly_->series(coeffs, maxdeg, (x - shift_)/scale_);
    }

    Matrix<double> OSDE1D::tMatrix(const unsigned maxdeg) const
    {
        Matrix<double> result(maxdeg+1U, maxdeg+1U);
        unsigned degs[2];
        for (unsigned i=0; i<=maxdeg; ++i)
        {
            degs[0] = i;
            for (unsigned j=0; j<=i; ++j)
            {
                degs[1] = j;
                result[i][j] = jointIntegral(degs, 2);
            }
        }
        for (unsigned i=0; i<=maxdeg; ++i)
            for (unsigned j=i+1; j<=maxdeg; ++j)
                result[i][j] = result[j][i];
        return result;
    }

    double OSDE1D::weight(const double x) const
    {
        return poly_->weight((x - shift_)/scale_)/scale_;
    }

    void OSDE1D::densityCoeffs(const AbsDistribution1D& distro,
                               const unsigned nIntegrationPoints,
                               double* coeffs, const unsigned maxdeg) const
    {
        assert(coeffs);
        require_identical_support("densityCoeffs", *this, distro);
        for (unsigned deg=0; deg<=maxdeg; ++deg)
        {
            OSDE1DHlp1 hlp(*this, deg);
            coeffs[deg] = distro.expectation(hlp, nIntegrationPoints);
        }
    }

    void OSDE1D::densityCoeffsVariance(const AbsDistribution1D& distro,
                                       const unsigned nIntegrationPoints,
                                       const double* coeffs, const unsigned maxdeg,
                                       const double expectedSampleSize,
                                       double* variances) const
    {
        assert(coeffs);
        assert(variances);
        validate_sample_size("densityCoeffsVariance", expectedSampleSize);
        require_identical_support("densityCoeffsVariance", *this, distro);
        for (unsigned deg=0; deg<=maxdeg; ++deg)
        {
            OSDE1DHlp2 hlp(*this, deg, deg, coeffs[deg], coeffs[deg]);
            variances[deg] = distro.expectation(hlp, nIntegrationPoints)/expectedSampleSize;
        }
    }

    npstat::Matrix<double>
    OSDE1D::densityCoeffsCovariance(const AbsDistribution1D& distro,
                                    const unsigned nIntegrationPoints,
                                    const double* coeffs, const unsigned maxdeg,
                                    const double expectedSampleSize) const
    {
        assert(coeffs);
        validate_sample_size("densityCoeffsCovariance", expectedSampleSize);
        require_identical_support("densityCoeffsCovariance", *this, distro);
        npstat::Matrix<double> result(maxdeg+1U, maxdeg+1U);
        for (unsigned deg=0; deg<=maxdeg; ++deg)
            for (unsigned k=0; k<=deg; ++k)
            {
                OSDE1DHlp2 hlp(*this, deg, k, coeffs[deg], coeffs[k]);
                result[deg][k] = distro.expectation(hlp, nIntegrationPoints);
            }
        for (unsigned deg=0; deg<=maxdeg; ++deg)
            for (unsigned k=deg+1; k<=maxdeg; ++k)
                result[deg][k] = result[k][deg];
        result /= expectedSampleSize;
        return result;
    }

    unsigned OSDE1D::optimalDegreeHart(const double* coeffs,
                                       const double* variances,
                                       const unsigned maxdeg,
                                       const double k)
    {
        assert(coeffs);
        assert(variances);

        long double sum = 0.0L, bestSum = DBL_MAX;
        unsigned bestM = 0;

        for (unsigned i=1U; i<=maxdeg; ++i)
        {
            sum += k*variances[i] - coeffs[i]*coeffs[i];
            if (sum < bestSum)
            {
                bestSum = sum;
                bestM = i;
            }
        }
        return bestM;
    }

    std::pair<double,double> OSDE1D::supportRegion(
        const AbsClassicalOrthoPoly1D& poly1d,
        const double leftLimit, const bool leftIsFromSample,
        const double rightLimit, const bool rightIsFromSample,
        const unsigned long nPoints)
    {
        if (leftLimit >= rightLimit) throw std::invalid_argument(
            "In npstat::OSDE1D::supportRegion: "
            "the left limit value must be less than the right limit value");
        if (!(leftIsFromSample || rightIsFromSample))
            return std::pair<double,double>(leftLimit, rightLimit);

        const unsigned nMin = leftIsFromSample && rightIsFromSample ? 2U : 1U;
        if (nPoints < nMin) throw std::invalid_argument(
            "In npstat::OSDE1D::supportRegion: insufficient sample size");

        const double xmin = poly1d.xmin();
        const double xmax = poly1d.xmax();
        const double halfstep = (xmax - xmin)/2.0/nPoints;

        if (leftIsFromSample && rightIsFromSample)
        {
            LinearMapper1d mp(xmin + halfstep, leftLimit,
                              xmax - halfstep, rightLimit);
            return std::pair<double,double>(mp(xmin), mp(xmax));
        }
        else if (leftIsFromSample)
        {
            LinearMapper1d mp(xmin + halfstep, leftLimit,
                              xmax, rightLimit);
            return std::pair<double,double>(mp(xmin), rightLimit);
        }
        else
        {
            LinearMapper1d mp(xmin, leftLimit,
                              xmax - halfstep, rightLimit);
            return std::pair<double,double>(leftLimit, mp(xmax));
        }
    }
}
