#include <sstream>

#include "npstat/stat/OrthoPolyGOFTest1D.hh"
#include "npstat/stat/Distributions1D.hh"

namespace npstat {
    double OrthoPolyGOFTest1D::analyticPValue(
        const double stat, unsigned long /* sz */) const
    {
        Gamma1D chisq(0.0, 2.0, nUnmasked_/2.0);
        return chisq.exceedance(stat);
    }

    std::string OrthoPolyGOFTest1D::shortName() const
    {
        const unsigned twodeg = 2U*maxdeg_;

        std::ostringstream os;
        os << "OP_" << twodeg;
        if (twodeg != nUnmasked_)
        {
            // Figure out the number of the last masked eigenvector
            unsigned lastMasked = twodeg;
            for (unsigned i=0; i<twodeg; ++i)
                if (mask_[i])
                    lastMasked = i;
            assert(lastMasked < twodeg);

            // Add mask info
            os << '_';
            for (unsigned i=0; i<=lastMasked; ++i)
            {
                if (mask_[i])
                    os << '1';
                else
                    os << '0';
            }
        }
        return os.str();
    }

    double OrthoPolyGOFTest1D::inverseExceedance(
        const double pvalue, const unsigned long sz,
        const double smin, const double smax) const
    {
        if (pvalue <= 0.0 || pvalue >= 1.0) throw std::invalid_argument(
            "In npstat::OrthoPolyGOFTest1D::inverseExceedance: "
            "p-value argument must be inside (0, 1) interval");

        if (pvalue > 1.0e-10)
        {
            Gamma1D chisq(0.0, 2.0, nUnmasked_/2.0);
            return chisq.quantile(1.0 - pvalue);
        }
        else
            return AbsUnbinnedGOFTest1D::inverseExceedance(
                pvalue, sz, smin, smax);
    }

    double OrthoPolyGOFTest1D::eigenUncertainty(const unsigned i) const
    {
        if (i >= invsqr_.nRows()) return 0.0;
        const long double normsq = sumOfSquares(invsqr_[i], invsqr_.nColumns());
        assert(normsq > 0.0L);
        return 1.0L/sqrtl(normsq);
    }

    std::vector<double> OrthoPolyGOFTest1D::covmatEigenvector(const unsigned i) const
    {
        if (i >= invsqr_.nRows()) throw std::invalid_argument(
            "In npstat::OrthoPolyGOFTest1D::covmatEigenvector: "
            "eigenvector number is out of range");
        const unsigned ncol = invsqr_.nColumns();
        std::vector<double> vec(invsqr_[i], invsqr_[i]+ncol);
        const double sigma = eigenUncertainty(i);
        for (unsigned i=0; i<ncol; ++i)
            vec[i] *= sigma;
        return vec;
    }

    std::vector<OrthoPolyGOFTest1D::EigenInfo> OrthoPolyGOFTest1D::eigenComponents(
        const unsigned i, const double threshold) const
    {
        const std::vector<double>& vec = covmatEigenvector(i);
        std::vector<EigenInfo> result;

        unsigned iRow = 0;
        for (unsigned i=0; i<=maxdeg_; ++i)
            for (unsigned k=i ? i : 1U; k<=maxdeg_; ++k, ++iRow)
                if (std::abs(vec[iRow]) > threshold)
                    result.push_back(EigenInfo(vec[iRow], UUPair(i, k)));
        assert(iRow == vec.size());
        return result;
    }
}
