#ifndef NPSTAT_UNBINNEDGOFTESTS1D_HH_
#define NPSTAT_UNBINNEDGOFTESTS1D_HH_

/*!
// \file UnbinnedGOFTests1D.hh
//
// \brief Some one-sample goodness-of-fit tests (KS, AD, CvM, Zhang's, etc)
//
// Author: I. Volobouev
//
// November 2020
*/

#include "npstat/stat/AbsUnbinnedGOFTest1D.hh"

namespace npstat {
    /** Kolmogorov-Smirnov test */
    class KSTest1D : public AbsUnbinnedGOFTest1D
    {
    public:
        inline explicit KSTest1D(const AbsDistribution1D& d)
            : AbsUnbinnedGOFTest1D(d) {}
        
        inline virtual ~KSTest1D() {}

        inline virtual std::string shortName() const {return "KS";}

        inline virtual double testStatistic(
            const double* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual double testStatistic(
            const float* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual bool hasAnalyticPValue() const {return true;}

        virtual double analyticPValue(double stat, unsigned long sz) const;

    private:
        template<typename Numeric>
        double testStat(const Numeric* data, unsigned long lenData,
                        bool dataIsSorted) const;
    };

    /** Anderson-Darling test */
    class ADTest1D : public AbsUnbinnedGOFTest1D
    {
    public:
        inline explicit ADTest1D(const AbsDistribution1D& d)
            : AbsUnbinnedGOFTest1D(d) {}
        
        inline virtual ~ADTest1D() {}

        inline virtual std::string shortName() const {return "AD";}

        inline virtual double testStatistic(
            const double* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual double testStatistic(
            const float* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual bool hasAnalyticPValue() const {return true;}

        virtual double analyticPValue(double stat, unsigned long sz) const;

    private:
        template<typename Numeric>
        double testStat(const Numeric* data, unsigned long lenData,
                        bool dataIsSorted) const;
    };

    /** Cramer-von Mises test */
    class CvMTest1D : public AbsUnbinnedGOFTest1D
    {
    public:
        inline explicit CvMTest1D(const AbsDistribution1D& d)
            : AbsUnbinnedGOFTest1D(d) {}
        
        inline virtual ~CvMTest1D() {}

        inline virtual std::string shortName() const {return "CvM";}

        inline virtual double testStatistic(
            const double* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual double testStatistic(
            const float* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual bool hasAnalyticPValue() const {return true;}

        virtual double analyticPValue(double stat, unsigned long sz) const;

    private:
        template<typename Numeric>
        double testStat(const Numeric* data, unsigned long lenData,
                        bool dataIsSorted) const;
    };

    /**
    // Jin Zhang's ZK test. For this and other Jin Zhang's tests, see
    // Jin Zhang, "Powerful Goodness-of-Fit Tests Based on the Likelihood
    // Ratio", Journal of the Royal Statistical Society. Series B
    // (Statistical Methodology), Vol. 64, No. 2 (2002), pp. 281-294.
    */
    class ZhangZKTest1D : public AbsUnbinnedGOFTest1D
    {
    public:
        inline explicit ZhangZKTest1D(const AbsDistribution1D& d)
            : AbsUnbinnedGOFTest1D(d) {}

        inline virtual ~ZhangZKTest1D() {}

        inline virtual std::string shortName() const {return "ZK";}

        inline virtual double testStatistic(
            const double* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual double testStatistic(
            const float* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

    private:
        template<typename Numeric>
        double testStat(const Numeric* data, unsigned long lenData,
                        bool dataIsSorted) const;
    };

    /** Jin Zhang's ZA test */
    class ZhangZATest1D : public AbsUnbinnedGOFTest1D
    {
    public:
        inline explicit ZhangZATest1D(const AbsDistribution1D& d)
            : AbsUnbinnedGOFTest1D(d) {}

        inline virtual ~ZhangZATest1D() {}

        inline virtual std::string shortName() const {return "ZA";}

        inline virtual double testStatistic(
            const double* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual double testStatistic(
            const float* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

    private:
        template<typename Numeric>
        double testStat(const Numeric* data, unsigned long lenData,
                        bool dataIsSorted) const;
    };

    /** Jin Zhang's ZC test */
    class ZhangZCTest1D : public AbsUnbinnedGOFTest1D
    {
    public:
        inline explicit ZhangZCTest1D(const AbsDistribution1D& d)
            : AbsUnbinnedGOFTest1D(d) {}

        inline virtual ~ZhangZCTest1D() {}

        inline virtual std::string shortName() const {return "ZC";}

        inline virtual double testStatistic(
            const double* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

        inline virtual double testStatistic(
            const float* data, const unsigned long sz, const bool b) const
            {return testStat(data, sz, b);}

    private:
        template<typename Numeric>
        double testStat(const Numeric* data, unsigned long lenData,
                        bool dataIsSorted) const;
    };
}

#include "npstat/stat/UnbinnedGOFTests1D.icc"

#endif // NPSTAT_UNBINNEDGOFTESTS1D_HH_
