#ifndef NPSTAT_NUMPYARRAYUTILS_HH_
#define NPSTAT_NUMPYARRAYUTILS_HH_

#include <cassert>
#include <stdexcept>

#include "Python.h"
#include "numpy/arrayobject.h"

namespace npstat {
    // This function will return Numpy array typecode or will
    // throw an exception if the input object is not a Numpy array
    inline int numpyArrayType(PyObject* input)
    {
        if (!(input && PyArray_Check(input))) throw std::invalid_argument(
            "In npstat::numpyArrayType: input object is not a Numpy array");
        return PyArray_TYPE((PyArrayObject*)input);
    }

    inline bool isNumpyShapeCompatible(PyObject* input,
                                       const unsigned *dataShape, const unsigned rank)
    {
        if (!(input && PyArray_Check(input))) throw std::invalid_argument(
            "In npstat::isNumpyShapeCompatible: input object is not a Numpy array");
        const int ndims = PyArray_NDIM((PyArrayObject*)input);
        if (ndims != static_cast<int>(rank))
            return false;
        assert(dataShape);
        for (int i=0; i<ndims; ++i)
        {
            const unsigned sz = PyArray_DIM((PyArrayObject*)input,i);
            if (sz != dataShape[i])
                return false;
        }
        return true;
    }

    inline PyObject* linearNumpyArrayElement(PyObject* a, const unsigned long index)
    {
        const int typenum = numpyArrayType(a);

        unsigned long aSize = 1;
        const int dim = PyArray_NDIM((PyArrayObject*)a);
        for (int i=0; i<dim; ++i)
            aSize *= static_cast<unsigned long>(PyArray_DIM((PyArrayObject*)a,i));
        if (index >= aSize)
            throw std::invalid_argument("In npstat::linearArrayElement: "
                                        "linear index out of range");
        PyObject* result = 0;
        if (typenum == NumpyTypecode<double>::code)
            result = Py_BuildValue("d", ((double*)PyArray_DATA((PyArrayObject*)a))[index]);
        else if (typenum == NumpyTypecode<float>::code)
        {
            const double d = ((float*)PyArray_DATA((PyArrayObject*)a))[index];
            result = Py_BuildValue("d", d);
        }
        else if (typenum == NumpyTypecode<int>::code)
            result = Py_BuildValue("i", ((int*)PyArray_DATA((PyArrayObject*)a))[index]);
        else if (typenum == NumpyTypecode<long int>::code)
            result = Py_BuildValue("l", ((long int*)PyArray_DATA((PyArrayObject*)a))[index]);
        else if (typenum == NumpyTypecode<long long>::code)
        {
            const long int l = ((long long*)PyArray_DATA((PyArrayObject*)a))[index];
            result = Py_BuildValue("l", l);
        }
        else if (typenum == NumpyTypecode<unsigned char>::code)
        {
            const int i = ((unsigned char*)PyArray_DATA((PyArrayObject*)a))[index];
            result = Py_BuildValue("i", i);
        }
        else
            throw std::invalid_argument("In npstat::linearArrayElement: "
                                        "unsupported input array type");
        return result;
    }

    double sumOfSquaredDifferences(PyObject* input1, PyObject* input2)
    {
        const int typed = NumpyTypecode<double>::code;
        const int typenum1 = numpyArrayType(input1);
        const int typenum2 = numpyArrayType(input2);
        if (typenum1 != typed || typenum2 != typed) throw std::invalid_argument(
            "Expected double numpy arrays as arguments");
        const int ndims1 = PyArray_NDIM((PyArrayObject*)input1);
        const int ndims2 = PyArray_NDIM((PyArrayObject*)input2);
        if (ndims1 != ndims2) throw std::invalid_argument("incompatible array ranks");
        unsigned long len = 1;
        for (int i=0; i<ndims1; ++i)
        {
            const unsigned sz1 = PyArray_DIM((PyArrayObject*)input1,i);
            const unsigned sz2 = PyArray_DIM((PyArrayObject*)input2,i);
            if (sz1 != sz2) throw std::invalid_argument("incompatible array shapes");
            len *= sz1;
        }
        const double* d1 = (double*)PyArray_DATA((PyArrayObject*)input1);
        const double* d2 = (double*)PyArray_DATA((PyArrayObject*)input2);
        long double sum = 0.0L;
        for (unsigned long i=0; i<len; ++i)
        {
            const double delta = *d1++ - *d2++;
            sum += delta*delta;
        }
        return sum;
    }
}

#endif // NPSTAT_NUMPYARRAYUTILS_HH_
